;;
;; $Id: partial-query.scm,v 1.3 2005/08/09 13:15:51 ben Exp $
;;
;; This module offer support for "partial" queries. These are SQL queries
;; that someone writes and then stores in some context. Later on,
;; the application can execute these queries by name.
;;
;; The goal of this module is to split up the writing and maintaining
;; of queries from the using of them.
;;

(require-library 'sql/query)
(require-library 'sql/jdbc)
(require-library 'sql/postgresql)
(require-library 'sisc/libs/srfi/srfi-45)
(require-library 'sisc/libs/srfi/srfi-13)
(require-library 'sisc/libs/srfi/srfi-1)
(require-library 'sisc/libs/srfi/srfi-9)

(module com/amazingmedia/db/partial-query
    ((make-partial-query     make-pq)
     partial-query/with-conn
     (partial-query/map      pq-sql pq-bindings pq-args->sql)
     (partial-query/let      pq-sql pq-bindings pq-args->sql))
     
  (import sql/query)
  (import sql/jdbc)
  (import sql/postgresql)
  (import srfi-1)
  (import srfi-45)
  (import srfi-13)
  (import record)
  (import hashtable)
  
  (define-struct pq (sql bindings args->sql))

  ;;
  ;; The constructor of a partial query. We use syntax here, because
  ;; we want to provide an easy'ish way to specify the sql.
  ;;
  ;; sql -- is in the format ("SELECT .." "..." "...").
  ;;        The parens act as a string-join, because we assume
  ;;        that our statements will always be over one line long.
  ;; binding -- a list of symbols which map to query results
  ;;
  (define-syntax make-partial-query
    (syntax-rules ()
      ((_ (sql ...) bindings)
       (make-pq (string-join (list sql ...) " ") bindings list))
      ((_ (sql ...) bindings converter)
       (make-pq (string-join (list sql ...) " ") bindings converter))))


  ;;
  ;; establish a block where a connection is active.
  ;; useful if you want to run a series of queries all
  ;; in the same transactional context.
  ;;
  (define-syntax partial-query/with-conn
    (syntax-rules ()
      ((_ conn body ...)
       (jdbc/call/tran 
        (jdbc/get-connection (data-source))
        (lambda (conn)
          body ...)))))

  ;;
  ;; Given a partial query, execute the body of the map once
  ;; for each result set. The variables provided in the binding
  ;; section will be bound to each column in the result set.
  ;;
  ;; query -- the partial query to execute
  ;; (args ...) -- arguments to pass with the query, filling in ?'s 
  ;;      in the query
  ;; (bindings ...) -- the variables which will be bound to
  ;;      results in each map iteration
  ;;
  (define-syntax partial-query/map
    (syntax-rules (connection)
      ((_ query (connection conn) (args0 ...) (binding0 ...) body ...)
       (let loop ((rs (apply sql/execute
                             (append (list conn
                                           (pq-sql query))
                                     ((pq-args->sql query) args0 ...)))))
         (cond ((null? (force rs)) '())
               (else
                (cons
                 (let* ((row (car (force rs)))
                        (col-index -1)
                        (col-val (lambda (binding)
                                   (set! col-index (+ col-index 1))
                                   (hashtable/get row (list-ref (pq-bindings query)
                                                                col-index)))))
                   
                   (let* ((binding0 (col-val 'binding0)) ...)
                     body ...))
                 (loop (cdr (force rs))))))))
      ((_ query (args0 ...) (binding0 ...) body ...)
       (partial-query/with-conn conn
         (partial-query/map query (connection conn)
                            (args0 ...) (binding0 ...) body ...)))))


  ;;
  ;; Just like squery/map-results, but assume we only want to bind
  ;; to the first row. Useful as a kind of let statement that knows
  ;; how to lookup just one row.
  ;;
  (define-syntax partial-query/let
    (syntax-rules (connection)
      ((_ query (connection conn) (args0 ...) (binding0 ...) body ...)
       (let ((rs (apply sql/execute (append (list conn (pq-sql query))
                                            ((pq-args->sql query) args0 ...)))))
         (cond ((null? (force rs))
                (let ((binding0 #f) ...)
                  body ...))
               (else
                   (let* ((row (car (force rs)))
                          (col-index -1)
                          (col-val (lambda (binding)
                                     (set! col-index (+ col-index 1))
                                     (let ((v (hashtable/get row (list-ref
                                                                  (pq-bindings query)
                                                                  col-index))))
                                       v))))
                     (let* ((binding0 (col-val 'binding0)) ...)
                       body ...))))))
      ((_ query (args0 ...) (binding0 ...) body ...)
       (partial-query/with-conn conn
         (partial-query/let query
             (connection conn)
             (args0 ...) (binding0 ...)
             body ...)))))

  )