;;
;; $Id: context-helper.scm,v 1.12 2005/08/09 13:15:51 ben Exp $
;;
;; We store lots and lots of stuff in our context. Which is convenient,
;; but can be a pain to keep track of. This module provides shortcuts to
;; dealing with stuff in our context. No new functionality is offered,
;; we are simply exposing functionality that can be found in our context.
;;

(require-library 'com/amazingmedia/util/context)
(require-library 'com/amazingmedia/web/history)
(require-library 'com/amazingmedia/web/direct-link)
(require-library 'com/amazingmedia/web/messages)
(require-library 'com/amazingmedia/web/item)
(require-library 'com/amazingmedia/web/domain-ml)
(require-library 'com/amazingmedia/web/partial-form)
(require-library 'com/amazingmedia/db/partial-query)

(module com/amazingmedia/sps/web/context-helper
    (make-initial-context
     render render-final
     use-form
     (let-query partial-query/let context/get)
     (map-query partial-query/map context/get)
     (with-conn partial-query/with-conn)
     add-error errors? get-errors clear-errors
     apply-nav
     logged-in? logged-in-as
     map-cart empty-cart
     add-history map-history clear-history has-history?
     has-direct-link? add-direct-link get-direct-link clear-direct-link
     reset-context)

  (import com/amazingmedia/util/context)
  
  (import com/amazingmedia/web/domain-ml)
  (import com/amazingmedia/web/partial-form)
  (import com/amazingmedia/db/partial-query)
  (import com/amazingmedia/web/item)
  (import com/amazingmedia/web/history)
  (import com/amazingmedia/web/messages)
  (import com/amazingmedia/web/direct-link)


  ;;
  ;; Return a fresh intialized context. This is where we define
  ;; what keys things are stored under, and other details such as 
  ;; that.
  ;;
  ;; returns a full loaded context
  ;;
  ;; queries-thunk -- the function to invoke to get the set of queries used by
  ;;     the system
  ;; layouts-thunk -- same as queries-thunk, but for layouts
  ;; forms-thunk -- same as queries-thunk, but for forms
  ;; nav-points-thunk -- same as queries-thunk, but for nav-points in the system
  ;;
  (define (make-initial-context queries-thunk layouts-thunk 
                                forms-thunk nav-points-thunk)
    (make-context 'immutable
                  `((cart         . ,(make-item-context))
                    (history      . ,(make-history-context))
                    (queries      . ,(make-consultant queries-thunk))
                    (layouts      . ,(make-consultant layouts-thunk))
                    (forms        . ,(make-consultant forms-thunk))
                    (nav-points   . ,(make-consultant nav-points-thunk))
                    (messages     . ,(make-messages-context '(warnings errors)))
                    (logged-in-as . #f))))

  ;;
  ;; Send some html back to the user. Make use of the layouts defined
  ;; in our context.
  ;;
  ;; ctx -- the context in use 
  ;; sxml -- the sxml to send to the user's browser
  ;;
  (define (render ctx sxml)
    (domain-ml/send ctx 'layouts sxml #f))

  ;;
  ;; Just like (render ...) but clear the continuation table for the
  ;; user, making the back button no longer useful.
  ;;
  ;; ctx -- the context in use 
  ;; sxml -- the sxml to send to the user's browser
  ;;
  (define (render-final ctx sxml)
    (domain-ml/send ctx 'layouts sxml #t))

  ;;
  ;; Use a form from our context and render it as HTML
  ;;
  ;; ctx -- the context in use
  ;; name -- the form name to use
  ;; success-k -- the function to invoke if the form passes validation
  ;; fail-k -- the function to invoke if the form can't be validated
  ;; args -- the list of arguments needed to populate the form
  ;; return the HTML of the form, ready to be displayed
  ;; 
  (define (use-form ctx name success-k fail-k . args)
    (domain-ml/transform
     ctx 'layouts
     (apply partial-form/use
            (append
             (list ctx 
                   (context/get `(forms ,name) ctx)
                   success-k fail-k)
             args))))

   
  ;;
  ;; Just like partial-query/let, but assume we know where our queries
  ;; are located
  ;;
  ;; ctx -- the context in use
  ;; name -- the name of the query
  ;; args -- the arguments to the query
  ;; bindings -- the variable names we should have bound in this block
  ;; body -- any sequence of statements
  ;;
  (define-syntax let-query
    (syntax-rules (connection)
      ((_ ctx name (args ...) (bindings ...) body ...)
       (partial-query/let (context/get '(queries name) ctx)
           (args ...)
           (bindings ...)
         body ...))
      ((_ ctx name  (connection conn) (args ...) (bindings ...) body ...)
       (partial-query/let (context/get '(queries name) ctx)
           (connection conn) (args ...) (bindings ...)
           body ...))))


  ;;
  ;; Just like partial-query/map, but we know where to find our 
  ;; queries
  ;;
  ;; ctx -- the context in use
  ;; name -- the name of the query
  ;; args -- arguments for the query
  ;; bindings -- the bindings of the arguments
  ;; body -- the body of the arguments
  ;;
  (define-syntax map-query
    (syntax-rules (connection)
      ((_ ctx name (args ...) (bindings ...) body ...)
       (partial-query/map 
           (context/get '(queries name) ctx)
           (args ...) (bindings ...)
         body ...))
      ((_ ctx name (connection conn) (args ...) (bindings ...) body ...)
       (partial-query/map (context/get '(queries name) ctx) (connection conn)
                          (args ...) (bindings ...)
                          body ...))))

  ;;
  ;; Establish a block where a connection can live
  ;;
  ;; conn -- the connection object
  ;;
  (define-syntax with-conn
    (syntax-rules ()
      ((_ conn body ...)
       (partial-query/with-conn conn body ...))))
            


  ;;
  ;; Return true if we have at least one error in our context.
  ;;
  (define (errors? ctx)
    (has-messages? '(messages errors) ctx))

  ;;
  ;; Add a message to our context if it's non-#f. This makes it easy
  ;; to avoid an extra conditional: just add the message, it's ok 
  ;; if it's #f.
  ;; 
  ;; ctx -- the context in use
  ;; message -- the message to add to our errors. If mesage is #f, we 
  ;;            do not add the error in.
  ;;
  ;; returns the updated context
  ;;
  (define (add-error ctx message)
    (if message
        (context/add '(messages errors) message ctx)
        ctx))

  ;;
  ;; Get the list of errors associated with this context
  ;;
  ;; ctx -- the context to check
  ;; return a list of errors, maybe null? if we have no errors
  ;;
  (define (get-errors ctx)
    (context/get '(messages errors) ctx))

  ;;
  ;; Clear out our errors object
  ;;
  (define (clear-errors ctx)
    (context/put '(messages errors) '() ctx))


  ;;
  ;; This is like scheme `apply', but looks up our functions
  ;; in the nav-points map
  ;;
  ;; ctx -- the context to use
  ;; name -- the name of the nav point
  ;; args -- arguments, which will have ctx prepeneded to,
  ;;         to apply name with
  ;;
  (define (apply-nav ctx name . args)
    (let ((fn (context/get `(nav-points ,name) ctx)))
      (if fn
          (apply fn (cons ctx args))
          (error (format (string-append "apply-nav failed: ~a does not "
                                        "correspond to a nav-point")
                         name)))))
      

  ;;
  ;; Are we currently logged in?
  ;;
  ;; ctx -- the context to check
  ;; return true if we are logged in, false otherwise
  ;;
  (define (logged-in? ctx)
    (context/get 'logged-in-as ctx))

  ;;
  ;; Who are we currently logged in as?
  ;;
  ;; ctx -- the context to check
  ;; return the username of the person we are logged in as, or
  ;;   #f if we aren't logged in
  ;;
  (define (logged-in-as ctx)
    (context/get 'logged-in-as ctx))

  ;;
  ;; Map over items in our shopping cart.
  ;;
  ;; ctx -- the context containing our cart
  ;; fn -- the function to be invoked with an item from the cart
  ;;
  (define (map-cart ctx fn)
    (context/map 'cart fn ctx))

  ;;
  ;; Empty out our cart. Maybe the items have been purchased?
  ;;
  (define (empty-cart ctx)
    (context/put 'cart (make-item-context) ctx))

  ;;
  ;; Add a new history element
  ;;
  ;; ctx -- the context in use
  ;; label -- name of the history
  ;; proc -- the proc to invoke when we have history
  ;; returns an updated context w/ the history added
  ;;
  (define (add-history ctx label proc)
    (context/add `(history ,label)
                 (make-history label proc)
                 ctx))

  ;;
  ;; Clear out our history
  ;;
  ;; ctx -- the context in use
  ;; return an updated history without our history entries
  ;;
  (define (clear-history ctx)
    (context/map 'history
                 (lambda (h)
                   (context/remove `(history ,(history-label h))
                                   ctx))
                 ctx)
    ctx)

  ;;
  ;; Does this context have any history in it, yet?
  ;;
  ;; ctx -- the context to check
  ;; return true if we have history of some kind
  ;;
  (define (has-history? ctx)
    (> (context/size 'history ctx) 0))

  ;;
  ;; map over our history
  ;;
  ;; ctx -- the context containing our history
  ;; proc -- the procedure to invoke with our history
  ;;
  (define (map-history ctx proc)
    (context/map 'history proc ctx))

  ;;
  ;; does this context have a direct link to this page?
  ;;
  ;; ctx -- the context to check
  ;; return true if we have a direct link
  (define (has-direct-link? ctx)
    (let ((found (context/get 'direct-link ctx)))
      (and found (not (null? found)))))

  ;;
  ;; add in a direct link to this to this page
  ;;
  ;; ctx -- the context in use
  ;; cmd -- the command to use
  ;; arg -- the argument to hand to the cmd
  ;; return the updated context
  ;; 
  (define (add-direct-link ctx cmd arg)
    (context/put 'direct-link 
                 (make-direct-link cmd (list arg))
                 ctx))

  ;;
  ;; get the current direct link in use
  ;;
  ;; ctx - the context to access
  ;; returns a pair of command and argument
  ;;
  (define (get-direct-link ctx)
    (context/get 'direct-link ctx))

  ;;
  ;; clear out the direct link
  ;;
  ;; ctx -- the context to clear out
  ;; returns the updated ctx
  ;;
  (define (clear-direct-link ctx)
    (context/put 'direct-link '() ctx))


  ;;
  ;; Reset the context for our next request.
  ;; This clears our transient things, like messages.
  ;;
  ;; ctx -- the context to clear
  ;; returns an updated, clean ctx
  ;;
  (define (reset-context ctx)
    ((compose clear-errors
              clear-direct-link)
     ctx))
  )