;;
;; $Id: context.scm,v 1.6 2005/08/15 15:42:13 ben Exp $
;;
;; Scheme code to represent a "context" -- a context is kind of an alist
;; and kind of a hashtable.  One one hand, they are used to store any keyed
;; data, like a hashtable, however they have the following special requirements:
;;
;; - updates are functional, all changes take in a context, don't touch the old
;;   context, but work on a new context
;;
;; - support for dealing with nested contexts, which makes it easy to access
;;   either a nested context or a nested value inside of a context
;;
;; One place where this comes in handy is handing around the state of a user
;; in a web app. As the user does things like log in, or add something to a
;; cart, their context gets updated. 
;;

(require-library 'sisc/libs/srfi/srfi-16)
(require-library 'com/amazingmedia/util/list)
;(require 'sort) ; XXX -- slib??

(module com/amazingmedia/util/context 
    (make-context
     context/get
     context/put
     context/add
     context/remove
     context/map
     context/size
     let-context
     make-consultant)

  (import srfi-1)
  (import srfi-16)
  (import record)
  (import com/amazingmedia/util/list)
  (import hashtable)

  ;;
  ;; What does a consultant do? Nothing of course. It is created
  ;; with a thunk, that when invoked will produce our needed value.
  ;;
  ;; It provides you with a way of having the computation of 
  ;; context be invoked every time, rather than storing the simple
  ;; value.
  ;;
  (define-struct consultant (thunk))

  (define-struct ctx (type data less-than equals resolver))

  ;;
  ;; Make a new context and return it.
  ;; If `values' is provided, it is assumed that it is a valid
  ;; alist made up of keys and values to be used to populate
  ;; our context
  ;;
  ;; The resolver is used in situations like this:
  ;; suppose you have a shopping-cart in a context
  ;; and you add an element to the cart. You would probably want
  ;; to take both elements and return a new element that was
  ;; (+ (item-quantity a) (item-quantity b))
  ;;
  ;;
  ;; values -- a list of cons pairs that represent the content of this context 
  ;; less-than -- optional the function used to sort entries
  ;;              while mapping through them. default value: #f, 
  ;;              don't sort. It will be asked to compare
  ;;              2 values: value1 value2 and should
  ;;              order them somehow
  ;; equals -- optional function used to tell if two elements in
  ;;           a context are the same, default: equal?
  ;; resolver -- take in an old and new objects and resolve them into one.
  ;;           used during the add process to support adding two different
  ;;           objects with the same key
  ;;
  (define make-context
    (case-lambda
     ((type) (make-context type '()))
     ((type values) (make-context type values #f))
     ((type values less-than) (make-context type values less-than equal?))
     ((type values less-than equal?) 
      (make-context type
                    values
                    less-than equal? 
                    (lambda (x y) (error "Add not supported: ~a, ~a"
                                         x y))))
     ((type values less-than equal? resolver)
      (make-ctx type
                (case type
                  ((immutable) values)
                  ((mutable) (alist->hashtable values equal?))
                  (else (error 'make-context "uknown type: [~a]" type)))  
                less-than equal? resolver))))

  ;;
  ;; Get an element from a context.
  ;;
  ;; key -- if key is a list? it is assumed
  ;;        that it represents a series of
  ;;        get requests. This makes deeply nested
  ;;        values available
  ;; ctx -- the context to look stuff up in
  ;;
  (define (context/get key ctx)
    (let loop ((k (normalize-key key))
               (ctx ctx))
      (cond ((not (ctx? ctx)) #f)
            ((null? (cdr k)) 
             (get (car k) ctx))
            (else
             (get (car k) ctx (lambda (found)
                                (loop (cdr k) found)))))))

  ;;
  ;; Replace an element in our context and return an updated
  ;; context. 
  ;;
  ;; key -- the key can be a list that describes which 
  ;;        element to update. 
  ;; val -- the value to insert
  ;; return an updated version of the entire context
  ;;
  (define (context/put key val ctx)
    (walk-context key 
                  (lambda (local-key local-ctx)
                    (put local-key val local-ctx))
                  ctx))

  ;;
  ;; Adding an element to a context is just like putting it,
  ;; however, if the element already exists, we are given a 
  ;; chance to reconcile the two elements and make a new one.
  ;;
  ;; key -- is the key of what we are looking for
  ;; val -- the new value to update
  ;; ctx -- the context to update
  ;; returns a new context that has been updated
  ;;
  (define (context/add key val ctx)
    (walk-context key 
                  (lambda (local-key local-ctx)
                    (let ((old-val (get local-key local-ctx)))
                      (put local-key (if old-val
                                         ((ctx-resolver local-ctx) old-val val)
                                         val) 
                           local-ctx)))
                  ctx))

  ;;
  ;; Remove an element from our context
  ;;
  ;; key -- the key, which may be in list form
  ;; ctx -- the context to update
  ;; the new context, minus the stuff
  ;;
  (define (context/remove key ctx)
    (walk-context key 
                  (lambda (local-key local-ctx)
                    (del local-key local-ctx))
                  ctx))

  ;;
  ;; Map over a given context, providing the keys and values to
  ;; proc
  ;;
  ;; key -- an optional argument, if provided, says which context
  ;;        to map over
  ;; proc -- will be called with the value in the context
  ;; ctx -- the context to map
  ;; returns the mapped list, not an updated context
  ;;
  (define context/map
    (case-lambda 
     ((proc ctx) (context/map '() proc ctx))
     ((key proc ctx)
      (let ((key (normalize-key key)))
        (define (sort list <) ; XXX add support for sorting again
          list)

        (cond ((null? key) (map (lambda (entry)
                                  (proc (cdrv entry)))
                                (if (ctx-less-than ctx)
                                    (sort (entries ctx)
                                          (ctx-less-than ctx))
                                    (entries ctx))))
              (else
               (get (car key) ctx 
                    (lambda (found)
                      (context/map (cdr key) proc found))
                    (lambda () '()))))))))

  ;;
  ;; Return the size of the context, that is the number of
  ;; entries in it.
  ;;
  ;; key -- an optional list of keys to describe which context
  ;;        we are talking about. If not provided, return the top
  ;;        level context
  ;; ctx -- the context to check
  ;; returns a count of the number of context entries
  ;;
  (define context/size
    (case-lambda
     ((ctx) (context/size '() ctx))
     ((key ctx)
      (let ((found (context/get key ctx)))
        (cond ((ctx? found) (size found))
              (else 
               (error "Context was expected, but found: ~a" found)))))))
  ;;
  ;; Syntax to allow easy access to our context. Use this like
  ;; any other let statement.
  ;;
  (define-syntax let-context
    (syntax-rules ()
      ((_ ((binding key) ...) ctx body ...)
       (let* ((binding (context/get key ctx)) ...)
         body ...))))
  

  (define (walk-context key action ctx)
    (let ((key (normalize-key key)))
      (cond ((null? (cdr key))
             (make-ctx (ctx-type ctx)
                       (action (car key) ctx)
                       (ctx-less-than ctx)
                       (ctx-equals ctx)
                       (ctx-resolver ctx)))
            (else
             (make-ctx
              (ctx-type ctx)
              (walk (car key)
                    (lambda (local-ctx)
                      (walk-context (cdr key) action local-ctx))
                    ctx)
              (ctx-less-than ctx)
              (ctx-equals ctx)
              (ctx-resolver ctx))))))

  (define (normalize-key key)
    (cond ((list? key) key)
          (else (list key))))

  ;;
  ;; given a value of a context, produce the value to 
  ;; return to the user
  ;;
  (define (->value obj)
    (cond ((consultant? obj) ((consultant-thunk obj)))
          (else obj)))

  (define (cdrv pair)
    (cond ((pair? pair) (->value (cdr pair)))
          (else pair)))



  ;;
  ;; implement our basic operations in terms of both
  ;; mutable and immutable data structures
  ;;

  (define get 
    (case-lambda
     ((key ctx) (get key ctx (lambda (x) x)))
     ((key ctx success-k) (get key ctx success-k (lambda () #f)))
     ((key ctx success-k fail-k)
      (case (ctx-type ctx)
        ((immutable) (cond ((assoc key (ctx-data ctx) (ctx-equals ctx)) =>
                            (lambda (found) 
                              (success-k (cdrv found))))
                           (else (fail-k))))
        ((mutable)
         (cond ((hashtable/get (ctx-data ctx) key) =>
                (lambda (found) (success-k (cdrv found))))
               (else (fail-k))))
        (else (error 'get "unknown ctx type: ~a" (ctx-type ctx)))))))

  (define (del key ctx)
    (case (ctx-type ctx)
      ((immutable)
       (alist-delete key (ctx-data ctx) (ctx-equals ctx)))
      ((mutable)
       (hashtable/remove! (ctx-data ctx) key)
       (ctx-data ctx))))

  (define (put key value ctx)
    (case (ctx-type ctx)
      ((immutable) 
       (alist-cons key value
                   (alist-delete key 
                                 (ctx-data ctx)
                                 (ctx-equals ctx))))
      ((mutable)
       (hashtable/put! (ctx-data ctx) key value)
       (ctx-data ctx))
      (else (error 'put "unknown ctx type: ~a" (ctx-type ctx)))))


  (define (entries ctx)
    (case (ctx-type ctx)
      ((immutable) (map cdrv (ctx-data ctx)))
      ((mutable) (hashtable/map (lambda (k v) v) (ctx-data ctx)))))

  (define (size ctx)
    (case (ctx-type ctx)
      ((immutable) (length (ctx-data ctx)))
      ((mutable) (hashtable/size (ctx-data ctx)))))
  
  (define (walk key found-k ctx)
    (case (ctx-type ctx)
      ((immutable) 
       (map (lambda (entry)
              (cond ((equal? (car entry) key)
                     (cons (car entry)
                           (found-k (cdrv entry))))
                    (else entry)))
            (ctx-data ctx)))
      ((mutable)
       (let ((found (hashtable/get (ctx-data ctx))))
         (found-k found)
         (ctx-data ctx)))
      (else (error 'walk "unknown type: ~a" (ctx-type ctx)))))

  )
