;;
;; $Id: form-pred.scm,v 1.5 2005/08/01 15:24:16 ben Exp $
;;
;; Scheme code to provide a bunch of common form related
;; predicates. Such as requiring a field to be filled in
;; or requiring a field to be a number.
;;
;; The procedures, when executed, produce a function which
;; will work as a form validator. See the partial-form file
;; for more info
;;
;;

(require-library 'com/amazingmedia/web/partial-form)
(require-library 'com/amazingmedia/util/context)
  
(module com/amazingmedia/web/form-pred
    (field-filled-in?
     field-is-a-number?
     field-is-a-positive-number?
     field-validates?
     fields-validate?
     try-all)

  (import com/amazingmedia/web/partial-form)
  (import com/amazingmedia/util/context)
  (import srfi-1)



  ;;
  ;; process a validator, calling fail-k and adding
  ;; messages, if the return value from the test 
  ;; suggests that's the right thing to do
  ;;
  (define (process ctx result success-k fail-k)
    ;; XXX -- this is app specific, shouldn't live here.
    (define (add-error ctx mesg)
      (context/add '(messages errors) mesg ctx))
    (cond ((string? result) 
           (fail-k (add-error ctx result)))
          ((list? result) 
           (let loop ((messages result) (ctx ctx))
             (cond ((null? messages) (fail-k ctx))
                   (else (loop (cdr messages)
                               (add-error ctx (car messages)))))))
          ((eq? result 'ok) (success-k ctx))
          (result (success-k ctx))
          (else (fail-k ctx))))
            

  ;;
  ;; This isn't a form predicate. However, it allows you
  ;; to group a bunch of predicates together and run them 
  ;; all. If anyone of them fails, the whole validation
  ;; will fail, however, all of the predicates will still
  ;; be run.
  ;;
  ;; This allows you to collect a bunch of error messages
  ;; and show them all at once to the user
  ;;
  ;; validators -- a list of validators to check
  ;;
  (define (try-all . validators)
    (make-validator 
     (apply append
            (map validator-field-names validators))
     (lambda (ctx success-k fail-k . args)
       (let loop ((validators validators)
                  (k  success-k)
                  (ctx ctx)
                  (args args))
         (cond ((null? validators) (k ctx))
               (else
                (let ((proc (validator-proc (car validators)))
                      (nargs  (length (validator-field-names (car validators)))))
                  (apply proc (append (list ctx
                                            (lambda (ctx)
                                              (loop (cdr validators)
                                                    k
                                                    ctx
                                                    (drop args nargs)))
                                            (lambda (ctx)
                                              (loop (cdr validators)
                                                    fail-k
                                                    ctx
                                                    (drop args nargs))))
                                      (take args nargs))))))))))


  ;;
  ;; is field-name filled in with some value?
  ;;
  (define (field-filled-in? field-name label)
    (make-validator (list field-name)
                    (lambda (ctx success-k fail-k value)
                      (process ctx 
                               (if (and value (not (equal? value "")))
                                   'ok
                                   (format "~a is a required field" label))
                               success-k fail-k))))
  ;;
  ;; is field-name a valid number?
  ;;
  (define (field-is-a-number? field-name label)
    (make-validator (list field-name)
                    (lambda (ctx success-k fail-k value)
                      (process ctx
                               (if (or (not value) (not (string->number value)))
                                   (format "~a must be a number" label)
                                   'ok)
                               success-k fail-k))))

  ;;
  ;; is field-name a valid positive number?
  ;;
  (define (field-is-a-positive-number? field-name label)
    (make-validator (list field-name)
                    (lambda (ctx success-k fail-k value)
                      (process ctx
                               (if (or (not value) (not (string->number value))
                                       (not (positive? (string->number value))))
                                   (format "~a must be a positive number" label)
                                   'ok)
                               success-k fail-k))))

  ;;
  ;; does field-name's value, when applied to the check-fn.
  ;; The first argument to the check-fn is always the ctx.
  ;;
  ;; The check-fn should return true if all is OK, or it can
  ;; return a string or list of strings which are stored
  ;; as error messages.
  ;;
  ;; return true?
  ;;
  (define (field-validates? field-name check)
    (make-validator (list field-name)
                    (lambda (ctx success-k fail-k value)
                      (process ctx (check value) success-k fail-k))))

  ;;
  ;; Same as field-validates, but take in a list of fields
  ;; and call the validation function with all those fields.
  ;; they will be called in the order in which they specified.
  ;;
  ;; The first argument to the check-fn is always the ctx.
  ;;
  ;; So if you say:
  ;;  (fields-validate '("foo" "bar") checker)
  ;; you'll be doing:
  ;;  (checker ctx <value-of-foo> <value-of-bar>)
  ;;
  (define (fields-validate? field-names check)
    (make-validator field-names
                    (lambda (ctx success-k fail-k . values)
                      (process ctx (apply check (cons ctx values))
                               success-k fail-k))))
  )
