;;
;; $Id: partial-form.scm,v 1.4 2005/08/01 15:24:16 ben Exp $
;;
;; Scheme code to develop a concept known as partial forms: these are like
;; partial queries -- except they relate to web forms (you know, those annoying
;; things you have to enter your data into).
;;
;; A web form, in this framework, has three components:
;; - rendering the form
;; - valdiating the form's data/converting the data from strings to valid types
;; - processing the form -- doing someting the form's values
;;
;; This framework allows you to specify the first two operations in one
;; location, and then use the form and handle processing in another location.
;;
;;
;; Rendering A Form -- the real rendering of the form is handled by the domain-ml
;; package, which will convert from the form description to HTML. However, the
;; the rendering handlers will be responsible for at least constructing 
;; the domain-ml specific form language
;;
;;

(require-library 'sisc/libs/srfi/srfi-9)

(module com/amazingmedia/web/partial-form
    (make-partial-form
     make-validator
     validator-field-names
     validator-proc
     partial-form/use)

  (import record)
  (import siscweb/bindings)
  (import com/amazingmedia/util/context)


  ;;
  ;; a validator is a function that gets 
  ;; invoked with the arguments:
  ;;  (v  ctx success-k fail-k . args)
  ;; where args are the values of the named
  ;; fields. the validator calls success-k or
  ;; fail-k depending on whether or not the
  ;; validation succeeded. both success-k and fail-k
  ;; take a single argument, an updated context
  ;;
  ;; field-names -- a list of field names to pull
  ;;   from the form bindings to validate
  ;; proc -- the procedure to call as above
  ;;
  (define-struct validator (field-names proc))

  ;;
  ;; A form is a renderer and validator, as well as a collection of elements.
  ;;
  ;; field-names -- the list of string field names used to generate
  ;;     an initial arguments list.
  ;; renderer -- a function that takes in a handler proc and current value
  ;;     and generates the SXML to generate this form 
  ;; validators -- a list validator objects
  ;; fields-process -- a function that takes in a list of the values
  ;;     of each the elements of he field-names list, and converts it
  ;;     to the argument list of the success-k and fail-k processes.
  ;;     For identity purposes, the function `list' can be passed in,
  ;;     though this is the ideal place to convert raw values into
  ;;     entities.
  ;;
  (define-struct partial-form (field-names 
                               renderer 
                               validators 
                               fields-processor))


  ;;
  ;; This procedure should be invoked when you are ready to "use" a form.
  ;; That is, have it rendered and say what actions should happen if the
  ;; form succeeds for fails.
  ;;
  ;; form -- the partial-form to render and use
  ;; success-k -- the function to invoke with a context and the arguments
  ;;              from the form.
  ;; fail-k -- invoked with the same arguments as success-k, but
  ;;           called when the form failed to validate
  ;; args -- the arguments to be passed to the form handler
  ;;
  (define (partial-form/use ctx form success-k fail-k . args)
    (define (validate-field validator ctx bindings success-k fail-k)
      (apply (validator-proc validator)
             (append (list ctx success-k fail-k)
                     (map (cut extract-single-binding <> bindings)
                          (validator-field-names validator)))))

    (define (validate ctx validators bindings success-k fail-k)
      (cond ((null? validators) (success-k ctx))
            (else
             (validate-field (car validators) ctx bindings
                             (lambda (ctx)
                               (validate ctx (cdr validators) bindings success-k fail-k))
                             fail-k))))

    (apply (partial-form-renderer form)
           (cons 
            (lambda (req)
              (let* ((bindings  (get-bindings req))
                     (arg-list (apply (partial-form-fields-processor form)
                                      (map (cut extract-single-binding <> bindings) 
                                           (partial-form-field-names form)))))
                ;; XXX -- shouldn't have '(messages errors) here, that's
                ;; app specific
                (validate (context/put '(messages errors) '() ctx)
                          (partial-form-validators form) bindings
                          (lambda (ctx)
                            (apply success-k (cons ctx arg-list)))
                          (lambda (ctx)
                            (apply fail-k (cons ctx arg-list))))))
            args)))


  )
