--
-- $Id: functions.sql,v 1.8 2005/08/01 15:24:16 ben Exp $
--

--
-- Given an item id -- form the item's name by concatenating
-- a bunch of columns on the record
--
CREATE OR REPLACE FUNCTION item_name(target_id Item.itemId%TYPE) RETURNS text AS $$
DECLARE
    item_rec  Item%ROWTYPE;
    prod_name text;
    name      text;
BEGIN
    SELECT INTO item_rec * FROM Item    WHERE itemId = target_id;
    SELECT INTO prod_name p.name 
    FROM 
      Product p, 
      Item    i
    WHERE
      p.productId = i.productId  AND
      i.itemId    = target_id;

    name := item_rec.attr1;
    IF item_rec.attr2 != NULL THEN
        name := name || ' ' || item_rec.attr2;
    END IF;
    IF item_rec.attr3 != NULL THEN
        name := name || ' ' || item_rec.attr3;
    END IF;
    IF item_rec.attr4 != NULL THEN
        name := name || ' ' || item_rec.attr4;
    END IF;
    IF item_rec.attr5 != NULL THEN
        name := name || ' ' || item_rec.attr4;
    END IF;

    RETURN name || ' ' || prod_name;
END
$$ LANGUAGE plpgsql;


--
-- Custom type: allows us to define verbose booleans, that is a boolean
-- value that caries along with it an explination of it's value
--
CREATE  TYPE vboolean AS (
   value  boolean,
   reason text
);

--
-- can this username and password log in? return yet/no and why.
--
CREATE OR REPLACE FUNCTION check_login(target_username Users.username%TYPE,
                                       target_password Users.password%TYPE) 
    RETURNS vboolean AS $$
DECLARE
    found_user Users;
    status     vboolean;
BEGIN
    SELECT INTO found_user * FROM Users WHERE username = target_username;
    IF found_user.userId IS NULL THEN
        SELECT INTO status false, 'username not found';
        RETURN status;
    END IF;

    IF found_user.password <> target_password THEN
        SELECT INTO status false, 'passwords do not match';
        RETURN status;
    END IF;
    
    SELECT INTO status true, 'ok';
    RETURN status;
END
$$ LANGUAGE plpgsql;

--
-- does this username exist?
--
CREATE OR REPLACE FUNCTION user_exists(target_username Users.username%TYPE)
    RETURNS boolean AS $$
DECLARE
    found_user Users;
    val        boolean;
BEGIN
    SELECT INTO found_user * FROM Users WHERE username = target_username;
    IF found_user.userId IS NULL THEN
        RETURN false;
    ELSE
        RETURN true;
    END IF;
END
$$ LANGUAGE plpgsql;

--
-- register a new user in the system
--
CREATE OR REPLACE FUNCTION register(target_username Users.username%TYPE,
                                    target_password Users.password%TYPE)
    RETURNS Users.userId%TYPE AS $$
DECLARE
    found_user Users;
    user_id    bigint;
BEGIN
    SELECT INTO found_user * FROM Users WHERE username = target_username;
    IF found_user.userId <> NULL THEN
        RAISE EXCEPTION 'user [%] already exists, can not register again', 
                found_user.userId;
    END IF;

    user_id := nextval('userIdSeq');
    INSERT INTO Users (userId, username, password) VALUES (
        user_id, target_username, target_password
    );
    RETURN user_id;
END
$$ LANGUAGE plpgsql;


--
-- purchase some stuff from our store
--
CREATE OR REPLACE FUNCTION purchase(
        target_username     Users.username%TYPE,
        cc_num              CreditCard.ccNum%TYPE,
        cc_type             CreditCard.cctype%TYPE,
        cc_exp_month        CreditCard.ccExpMonth%TYPE,
        cc_exp_year         CreditCard.ccExpYear%TYPE,
        bill_fname          Address.firstName%TYPE,
        bill_lname          Address.lastName%TYPE,
        bill_street1        Address.street1%TYPE,
        bill_street2        Address.street2%TYPE,
        bill_city           Address.city%TYPE,
        bill_state          Address.state%TYPE,
        bill_zip            Address.zip%TYPE,
        ship_fname          Address.firstName%TYPE,
        ship_lname          Address.lastName%TYPE,
        ship_street1        Address.street1%TYPE,
        ship_street2        Address.street2%TYPE,
        ship_city           Address.city%TYPE,
        ship_state          Address.state%TYPE,
        ship_zip            Address.zip%TYPE
    )
    RETURNS bigint AS $$
DECLARE
    found_user      Users;
    order_id        bigint;
    bill_addr_id    bigint;
    ship_addr_id    bigint;
    cc_id           bigint;
BEGIN
    SELECT INTO found_user * FROM Users WHERE username = target_username;
    SELECT INTO order_id nextval('orderIdSeq');
    SELECT INTO bill_addr_id nextval('addressIdSeq');
    SELECT INTO ship_addr_id nextval('addressIdSeq');
    SELECT INTO cc_id nextval('ccIdSeq');

    IF found_user.userId IS NULL
    THEN
        RAISE EXCEPTION 'purchase failed, username [%] does not exist', target_username;
    END IF;

    INSERT INTO Address (addressId, firstName, lastName, 
                         street1, street2, city, state, zip) VALUES (
        bill_addr_id, bill_fname, bill_lname, bill_street1, bill_street2,
        bill_city, bill_state, bill_zip
    );

    INSERT INTO Address (addressId, firstName, lastName, 
                         street1, street2, city, state, zip) VALUES (
        ship_addr_id, ship_fname, ship_lname, ship_street1, ship_street2,
        ship_city, ship_state, ship_zip
    );

    INSERT INTO CreditCard (ccId, ccNum, ccType, ccExpMonth, ccExpYear) VALUES (
        cc_id, cc_num, cc_type, cc_exp_month, cc_exp_year
    );

    INSERT INTO Orders (orderId, userId, orderDate, ccId, 
                        billAddrId, shipAddrId, status) VALUES (
        order_id, found_user.userId, now(), cc_id, 
        bill_addr_id, ship_addr_id, 'Pending'
    );

    RETURN order_id;
END;
$$ LANGUAGE plpgsql;

--
-- add some line items in with a purchased order
--
CREATE OR REPLACE FUNCTION add_line_item(
        order_id            Orders.orderId%TYPE,
        item_id             Item.itemId%TYPE,
        num                 OrderLineItem.quantity%TYPE,
        price               OrderLineItem.unitCost%TYPE
    ) RETURNS OrderLineItem.lineNum%TYPE AS $$
DECLARE
    line_num    integer;
    available   Inventory;
BEGIN
    SELECT INTO available * FROM Inventory WHERE itemId = item_id;
    IF available.quantity < num THEN
        RAISE EXCEPTION 'requested purchase of % %, but only % left', 
            num, item_id, available.quantity;
    END IF;

    UPDATE Inventory SET quantity = quantity - num 
        WHERE itemId = item_id;
    SELECT INTO line_num COUNT(*) FROM OrderLineItem WHERE orderId = order_id;
    INSERT INTO OrderLineItem (orderId, lineNum, itemId, quantity, unitCost) VALUES (
                               order_id, line_num, item_id, num, price
    );
    RETURN line_num;
END;
$$ LANGUAGE plpgsql;

--
-- get a summary view of a user's orders
--
CREATE TYPE order_summary_line AS (
    orderId                 bigint,
    shipTo                  text,
    total                   numeric(20,2),
    purchaseDate            timestamp    
);

CREATE OR REPLACE FUNCTION order_summary(username Users.username%TYPE)
    RETURNS SETOF order_summary_line AS $$
DECLARE
    total       numeric(20,2);
    line        order_summary_line;
BEGIN
    FOR line IN SELECT 
                     o.orderId, 
                     a.firstName || ' ' || a.lastName, 
                     sum(l.unitCost), 
                     o.orderDate
                 FROM
                     Orders        o,
                     Users         u,
                     Address       a,
                     OrderLineItem l
                 WHERE
                     o.userId    = u.userId      AND
                     u.username  = username      AND
                     a.addressId = o.shipAddrId  AND
                     l.orderId   = o.orderId
                 GROUP BY o.orderId, a.firstName, a.lastName, o.orderDate
                 ORDER BY o.orderDate DESC
                LOOP
        RETURN NEXT line;
    END LOOP;
    RETURN;
END;
$$ LANGUAGE plpgsql;


--
-- search procedures
--
CREATE TYPE search_result AS (
    itemId                  textId,
    productId               textId,
    name                    text,
    description             text,
    image                   text
);

CREATE OR REPLACE FUNCTION search(query text) 
    RETURNS SETOF search_result AS $$
DECLARE
    result         search_result;
    q              text;        
BEGIN
    q := lower('%' || query || '%');
    FOR result IN 
        SELECT 
            i.itemId,
            p.productId,
            item_name(i.itemId) AS name,
            i.description,
            i.image
        FROM
            Item        i,
            Product     p,
            Category    c
        WHERE
            i.productId = p.productId           AND
            p.categoryId = c.categoryId         AND
            (lower(item_name(i.itemId)) LIKE  q   OR
             lower(i.description)       LIKE  q   OR
             lower(p.name)              LIKE  q   OR
             lower(p.description)       LIKE  q   OR
             lower(c.name)              LIKE  q)     LOOP
        RETURN NEXT result;
    END LOOP;
    RETURN;
END;
$$ LANGUAGE plpgsql;
